from __future__ import absolute_import

import fnmatch
import logging
import os
import re

from .match import NotMatching, match_from_comment

logger = logging.getLogger(__name__)


REGEX = re.compile(r'# Raincoat: ([a-z]+) (.+)(\n|#|$)')
ARGS_REGEX = re.compile(r' *([^ ]+): *([^ ]+)(?: *|$)')


def find_in_string(file_content, filename):
    for match in REGEX.finditer(file_content):
        lineno = lineno = file_content.count(
            os.linesep, 0, match.start()) + 1

        kwargs_str = match.group(2).strip()
        kwargs = dict(
            pair.groups()
            for pair in ARGS_REGEX.finditer(kwargs_str))

        try:
            match = match_from_comment(match_type=match.group(1),
                                       filename=filename,
                                       lineno=lineno,
                                       **kwargs)

        except NotMatching:
            logger.warning("Unrecognized Raincoat comment at {}:{}\n{}".format(
                filename, lineno, match.group(0)))
            continue

        yield match


def find_in_file(filename):
    with open(filename) as handler:
        return find_in_string(handler.read(), filename)


def list_python_files(base_dir=".", exclude=None):
    exclude = exclude or []
    exclude = {os.path.normpath(path) for path in exclude}
    for root, folders, files in os.walk(base_dir, topdown=True):

        # Prune excluded folders
        full_pathes = [os.path.normpath(os.path.join(root, folder))
                       for folder in folders]

        folders_to_remove = {
            os.path.basename(folder)
            for pattern in exclude
            for folder in fnmatch.filter(full_pathes, pattern)}

        if folders_to_remove:
            folders[:] = set(folders) - folders_to_remove

        # Prune excluded files
        full_pathes = [os.path.normpath(os.path.join(root, file))
                       for file in files]

        files_to_remove = {os.path.basename(file)
                           for pattern in exclude
                           for file in fnmatch.filter(full_pathes, pattern)}

        for file in set(files) - files_to_remove:
            if file.endswith(".py"):
                yield os.path.normpath(os.path.join(root, file))


def find_in_dir(base_dir=".", exclude=None):
    for python_file in list_python_files(base_dir, exclude=exclude):
        for match in find_in_file(python_file):
            yield match



from __future__ import absolute_import

import os
import tarfile
import zipfile

from distutils.version import StrictVersion
import pip
import pkg_resources
import requests

from raincoat.constants import FILE_NOT_FOUND
from raincoat import github_utils


def download_package(package, version, download_dir):
    full_package = "{}=={}".format(package, version)

    exit_code = pip.main(
        ["download", "--no-deps", "-d", download_dir, full_package])
    if exit_code != 0:
        raise ValueError("Error while fetching {} via pip.".format(
            full_package))


def open_in_wheel(wheel, pathes):
    with zipfile.ZipFile(wheel, 'r') as zf:
        sources = {}
        for path in pathes:
            try:
                source = zf.open(path, 'r').read().decode("UTF-8")
            except KeyError:
                source = FILE_NOT_FOUND
            sources[path] = source
        return sources


def open_in_tarball(tarball, pathes):
    sources = {}
    with tarfile.open(tarball, 'r:gz') as tf:
        for path in pathes:
            top_level_dir = tf.next().name
            try:
                handler = tf.extractfile(os.path.join(top_level_dir, path))
                source = handler.read().decode("UTF-8")
            except KeyError:
                source = FILE_NOT_FOUND
            sources[path] = source
    return sources


def open_downloaded(download_path, pathes):
    archive_name, = os.listdir(download_path)

    archive_path = os.path.join(download_path, archive_name)
    ext = os.path.splitext(archive_name)[1]

    if ext == ".gz":
        return open_in_tarball(archive_path, pathes)
    elif ext == ".whl":
        return open_in_wheel(archive_path, pathes)
    else:
        raise NotImplementedError("Unrecognize archive format {}".format(ext))


def open_installed(installed_path, pathes):
    sources = {}
    for path in pathes:
        try:
            source = open(os.path.join(installed_path, path)).read()
        except IOError:
            source = FILE_NOT_FOUND
        sources[path] = source

    return sources


def get_current_or_latest_version(package):
    try:
        return True, pkg_resources.get_distribution(package).version
    except pkg_resources.DistributionNotFound:
        pass
    pypi_url = "http://pypi.python.org/pypi/{}/json".format(package)
    releases = requests.get(pypi_url).json()["releases"]

    versions = []

    for version in releases:
        try:
            parsed_version = StrictVersion(version)
        except ValueError:
            continue

        if parsed_version.prerelease:
            continue

        versions.append((parsed_version, version))

    return False, next(iter(sorted(versions, reverse=True)))[1]


def get_current_path(package):
    return pkg_resources.get_distribution(package).location


def get_branch_commit(repo, branch):
    # This may fail, but so far, I don't really know how.
    url = "https://api.github.com/repos/{}/branches/{}".format(repo, branch)
    with github_utils.get_session() as session:
        response = session.get(url)
        response.raise_for_status()
        return response.json()["commit"]["sha"]


def download_files_from_repo(repo, commit, files):
    result = {}

    template = "https://raw.githubusercontent.com/{}/{}/{}"
    with github_utils.get_session() as session:
        for file in files:
            url = template.format(repo, commit, file)
            response = session.get(url)
            if response.status_code != 200:
                content = FILE_NOT_FOUND
            else:
                content = response.text

            result[file] = content

    return result



import difflib
from collections import OrderedDict
from itertools import groupby
from operator import itemgetter

from raincoat import parse
from raincoat import constants
from raincoat.match import Match


class PythonChecker(object):
    """
    Find the matches related to the same source, find all the
    files it needs, then parse these files and compare the elements
    inside.

    This is an abstract class.
    """
    def current_source_key(self, match):
        """
        Should return a hashable identifier of the source code
        to read for the "current" version of the code.
        """
        raise NotImplementedError

    def match_source_key(self, match):
        """
        Should return a hashable identifier of the source code
        to read for the version of the code at the time the
        Raincoat comment (the "match") was written.
        """
        raise NotImplementedError

    def get_source(self, key, files):
        """
        The key is the one defined in *_source_key
        The files are the set of files we want in this source.
        """
        raise NotImplementedError

    def check(self, matches):
        """
        Main entrypoint
        """
        matches = list(matches)

        # A set of all source keys.
        # source_key should be: (source, path, element)
        source_keys = sorted(set(self.get_source_keys(matches)))

        # A dict: (source_key, path, element_key) > element
        elements = dict(self.get_elements(source_keys))

        # A generator yielding every match and the error message
        return self.run_matches(matches, elements)

    def get_source_keys(self, matches):
        for match in matches:
            path, element = match.get_path(), match.get_element()
            match_key = self.match_source_key(match)
            yield match_key, path, element
            current_key = self.current_source_key(match)
            yield current_key, path, element

    def get_elements(self, source_keys):
        grouped_keys = group_composite(source_keys)
        for source_key, files_dict in grouped_keys.items():
            files_source = self.get_source(source_key, set(files_dict))
            for path, element_names in files_dict.items():
                file_source = files_source[path]
                if file_source is constants.FILE_NOT_FOUND:
                    for element_name in element_names:
                        full_key = (source_key, path, element_name)
                        yield full_key, constants.FILE_NOT_FOUND
                else:
                    elements = parse.find_elements(
                        file_source, element_names)
                    for element_name, element_source in elements:
                        full_key = (source_key, path, element_name)
                        yield full_key, element_source

    def run_matches(self, matches, elements):
        for match in matches:
            path, element = match.get_path(), match.get_element()

            match_key = (self.match_source_key(match), path, element)
            current_key = (self.current_source_key(match), path, element)

            match_element = elements.get(match_key)
            current_element = elements.get(current_key)

            if match_element != current_element:
                yield self.run_match(
                    match=match,
                    match_key=match_key, match_element=match_element,
                    current_key=current_key, current_element=current_element)

    def run_match(self, match, match_key, match_element,
                  current_key, current_element):

        if match_element is constants.FILE_NOT_FOUND:
            return ("Invalid Raincoat PyPI comment : {match.path} does "
                    "not exist"
                    "".format(match=match),
                    match)
        if match_element is constants.ELEMENT_NOT_FOUND:
            return ("Invalid Raincoat PyPI comment : {match.element} does "
                    "not exist in {match.path}"
                    "".format(match=match),
                    match)

        if current_element is constants.FILE_NOT_FOUND:
            return ("File {match.path} disappeared"
                    "".format(match=match),
                    match)
        if current_element is constants.ELEMENT_NOT_FOUND:
            return ("Element {match.element} disappeared from "
                    "{match.path}"
                    "".format(match=match),
                    match)

        path = match.get_path()
        diff = "\n".join(difflib.unified_diff(
            match_element,
            current_element,
            fromfile=path, tofile=path,
            lineterm=""))
        return "Code is different:\n{}".format(diff), match


def group_composite(elements):
    """
    >>> group_composite(["abc", "abd", "acd", "bbc"])
    {"a": {"b": ["c", "d"], "c": ["d"]}, "b": {"b": ["c"]}}

    Input is expected to be sorted.
    All input is expected to be of the same length
    """
    elements = list(elements)

    # End of recursion
    if len(elements[0]) == 1:
        return list(map(itemgetter(0), elements))

    result = OrderedDict()
    for key, sub_elements in groupby(elements, itemgetter(0)):
        result[key] = group_composite(el[1:] for el in sub_elements)

    return result


class PythonMatch(Match):
    def get_path(self):
        return self.path

    def get_element(self):
        return self.element

    def format_line(self, line, color, i):
        line = super(PythonMatch, self).format_line(line, color, i)
        if line[0] in "+-@":
            line = color["diff" + line[0]](line)

        return line



"""
Responsible for gluing together all the parts of the package.
raincoat is the entrypoint that receives a path and
will:
 - Find all the Raincoat comments
 - Find all the corresponding code
 - Format errors
 - Yield them

"""

from __future__ import absolute_import

import itertools

from . import grep
from .match import check_matches
from .color import get_color


def class_key(match):
    return match.__class__


def class_key_name(match):
    return match.__class__.__name__


def raincoat(path, exclude=None, color=False):
    """
    Main entrypoint
    """
    matches = sorted(grep.find_in_dir(path, exclude=exclude),
                     key=class_key_name)

    matches_dict = {}
    for match_class, matches_for_class in itertools.groupby(
            matches, key=class_key):
        matches_for_class = list(matches_for_class)
        matches_dict[match_class.match_type] = matches_for_class

    color_obj = get_color(color)
    for error, match in check_matches(matches_dict):
        yield match.format(error, color_obj)



from itertools import count


class NotMatching(Exception):
    pass


class Match(object):
    # Filled at the end of the module
    match_type = NotImplemented
    checker = NotImplemented

    def __init__(self, filename, lineno):
        self.filename = filename
        self.lineno = lineno

    def __str__(self):
        return "Match in {}:{}".format(self.filename, self.lineno)

    def format(self, message, color):
        message = message.strip()
        result = ""

        result += color["match"](str(self)) + "\n"

        lines = message.splitlines()
        counter = count()

        for line in lines:
            line = line.strip()
            if line:
                result += self.format_line(line, color, next(counter))
                result += "\n"

        return result

    def format_line(self, line, color, i):
        if i == 0:
            line = color["message"](line)
        return line


def match_from_comment(match_type, filename, lineno, **kwargs):
    """
    Indentifies the correct Match subclass and
    creates a match
    """
    try:
        return match_types[match_type](filename, lineno, **kwargs)
    except KeyError:
        raise NotMatching


def check_matches(matches):
    for match_type, matches_for_type in matches.items():
        match_class = match_types[match_type]
        checker = match_class.checker

        if checker is NotImplemented:
            raise NotImplementedError(
                "{} has no checker".format(match_class))

        for difference in checker().check(matches_for_type):
            yield difference


def fill_match_types(match_types, match_classes):
    for match_class in match_classes:
        match_type = match_class.match_type

        if match_type is NotImplemented:
            raise NotImplementedError(
                "{} has no match_type".format(match_class))

        match_types[match_type] = match_class


from .pypi import PyPIMatch  # noqa
from .django import DjangoMatch  # noqa
from .pygithub import PyGithubMatch  # noqa
match_classes = [PyPIMatch, DjangoMatch, PyGithubMatch]

match_types = {}

fill_match_types(match_types, match_classes)



from __future__ import absolute_import

import sys

import click

from .glue import raincoat

__version__ = "0.8.0"

CONTEXT_SETTINGS = {'help_option_names': ['-h', '--help']}


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('path', nargs=-1, type=click.Path(exists=True))
@click.option('--exclude', '-e', multiple=True,
              help="Files and folders to exclude (e.g. 'test_*')")
@click.option('-c/-nc', '--color/--no-color', default=None,
              help="Should output be colorized ? (default : yes for TTYs)")
def main(path, exclude=None, color=True):
    """
    Analyze your code to find outdated copy-pasted snippets.
    "Raincoat has you covered when your code is not DRY."

    Full documentation at http://raincoat.readthedocs.io/en/latest/
    """
    if not path:
        path = ["."]

    if color is None:
        color = sys.stdout.isatty()

    errors = (error_match
              for element in path
              for error_match in raincoat(path=element,
                                          exclude=exclude,
                                          color=color))
    has_errors = False
    for line in errors:
        has_errors = True
        click.echo(line)

    sys.exit(int(has_errors))



import os

import requests


def get_session():
    session = requests.Session()
    token = os.getenv("RAINCOAT_GITHUB_TOKEN")
    if token:
        session.auth = tuple(token.split(":"))
    return session



import os
import shutil
import tempfile


class Cleaner(object):
    """
    Context manager that takes care of deleting
    objects when it goes out of scope.
    """
    def __init__(self):
        self.elements = set()

    def __enter__(self):
        return self

    def add(self, path):
        self.elements.add(path)
        return path

    def mkdir(self):
        return self.add(tempfile.mkdtemp())

    def __exit__(self, *args, **kwargs):
        for element in self.elements:
            if os.path.exists(element):
                if os.path.isdir(element):
                    shutil.rmtree(element)
                else:
                    os.remove(element)



from colorama import Fore, Style, init as colorama_init


colorama_init()

COLOR_DICT = {
    "neutral": Style.RESET_ALL,
    "match": Fore.YELLOW + Style.BRIGHT,
    "diff@": Fore.CYAN + Style.BRIGHT,
    "diff+": Fore.GREEN,
    "diff-": Fore.RED,
    "message": Fore.WHITE + Style.BRIGHT,
}
SEPARATOR = "=" * 80


class Color(object):
    def __init__(self, color_dict):
        self.color_dict = color_dict

    def get(self, name):
        return self.color_dict.get(name, "")

    def __getitem__(self, name):
        def apply_color(content):
            return self.get(name) + content + self.get("neutral")
        return apply_color


def get_color(color=False):
    color_dict = {}
    if color:
        color_dict = COLOR_DICT

    return Color(color_dict)



from collections import namedtuple

from raincoat import source
from raincoat.match import NotMatching
from raincoat.match.python import PythonChecker, PythonMatch
from raincoat.utils import Cleaner


PyPIKey = namedtuple("PyPIKey", "package version installed")


class PyPIChecker(PythonChecker):

    def __init__(self, *args, **kwargs):
        super(PyPIChecker, self).__init__(*args, **kwargs)
        self.package_cache = {}

    def current_source_key(self, match):
        if match.package in self.package_cache:
            key = self.package_cache[match.package]
            match.other_version = key.version
            return key

        installed, version = source.get_current_or_latest_version(
            match.package)
        pypi_key = PyPIKey(match.package, version, installed)

        self.package_cache[match.package] = pypi_key
        match.other_version = version

        return pypi_key

    def match_source_key(self, match):
        return PyPIKey(match.package, match.version, installed=False)

    def get_source(self, key, files):
        if key.installed:
            path = source.get_current_path(key.package)
            return source.open_installed(path, files)
        else:
            with Cleaner() as cleaner:
                path = cleaner.mkdir()
                source.download_package(key.package, key.version, path)
                return source.open_downloaded(path, files)


class PyPIMatch(PythonMatch):

    def __init__(self, filename, lineno, package, path, element=""):
        try:
            self.package, self.version = package.strip().split("==")
        except ValueError:
            raise NotMatching
        self.path = path.strip()
        self.element = element.strip()

        # This may be filled manually later.
        self.other_version = None

        super(PyPIMatch, self).__init__(filename, lineno)

    def __str__(self):
        return (
            "{match.package} == {match.version}{vs_match} "
            "@ {match.path}:{element} "
            "(from {match.filename}:{match.lineno})".format(
                match=self,
                vs_match=" vs {}".format(self.other_version)
                         if self.other_version else "",
                element=self.element or "whole module"))

    checker = PyPIChecker
    match_type = "pypi"



import re

import six

from raincoat.match import Match, NotMatching
from raincoat import source
from raincoat import github_utils


def get_merge_commit_sha1(ticket, session):

    # This is an adaptation of
    # https://github.com/django/code.djangoproject.com/blob/
    # cad96e2d980fc0453b34dd3d17ce6cb895e1aa89/trac-env/htdocs/
    # tickethacks.js#L99-L209

    # This is definitely the place where we could USE raincoat
    # ... But there's not yet raincoat comments for github
    # and this code is not released in PyPI
    ticket = int(ticket)
    pr_title_patterns = ["#{} ", "#{},", "#{}:", "#{})"]
    url = "https://api.github.com/search/issues?q="
    args = "repo:django/django+state:closed+in:title+type:pr+"
    args += "+".join(pattern.format(ticket)
                     for pattern in pr_title_patterns)
    response = session.get(
        url + six.moves.urllib.parse.quote(args, safe="+:"))
    response.raise_for_status()
    search_results = response.json()

    merged_regex = re.compile(r'(?:merged|fixed) in \b([0-9a-f]{6,40})\b')

    for pr in search_results["items"]:
        number = pr["number"]

        if number == ticket:
            # skip this element if PR id == ticket id
            continue

        merged = session.get(
            "https://api.github.com/repos/django/django/pulls/{}/merge"
            .format(number)).status_code == 204

        if merged:
            response = session.get(
                    "https://api.github.com/repos/django/django/pulls/{}"
                    .format(number))
            response.raise_for_status()
            pr_details = response.json()

            return pr_details["merge_commit_sha"]

        # Check if the PR was merged manually
        response = session.get(
            "https://api.github.com/repos/django/django/issues/{}/comments"
            .format(number))
        response.raise_for_status()
        comments = response.json()
        for comment in comments:
            merged_in = merged_regex.search(comment["body"])
            if merged_in:
                return merged_in.group(1)


def is_commit_in_version(commit, version, session):
    url = (
        "https://api.github.com/repos/django/django/compare/{}...{}"
        "".format(commit, version))
    response = session.get(url)
    response.raise_for_status()
    diff = response.json()
    return not diff.get("status") == "diverged"


class DjangoChecker(object):

    def get_match_info(self, matches):
        info = {}
        for match in matches:
            info.setdefault(match.ticket, []).append(match)
        return info

    def check(self, matches):
        __, django_version = source.get_current_or_latest_version("django")

        match_info = self.get_match_info(matches)

        return self.check_matches(match_info, django_version)

    def check_matches(self, match_info, django_version):
        with github_utils.get_session() as session:
            for ticket, ticket_matches in match_info.items():
                sha1 = get_merge_commit_sha1(ticket, session)
                if sha1:
                    if is_commit_in_version(sha1, django_version, session):
                        for match in ticket_matches:
                            yield (
                                "Ticket #{} has been merged in Django {}"
                                .format(ticket, django_version),
                                match)


class DjangoMatch(Match):
    checker = DjangoChecker
    match_type = "django"
    ticket_regex = re.compile(r'^#?(\d+)$')

    def __init__(self, filename, lineno, ticket):
        super(DjangoMatch, self).__init__(filename, lineno)

        try:
            regex_match = self.ticket_regex.match(ticket)
            self.ticket = int(regex_match.group(1).strip())
        except (AttributeError, TypeError):
            raise NotMatching

    def __str__(self):
        return (
            "Django ticket #{match.ticket} "
            "(from {match.filename}:{match.lineno})".format(match=self))



from collections import namedtuple

from raincoat import source
from raincoat.match import NotMatching
from raincoat.match.python import PythonChecker, PythonMatch


PyGithubKey = namedtuple("PyGithubKey", "repo commit")


class PyGithubChecker(PythonChecker):

    def __init__(self, *args, **kwargs):
        super(PyGithubChecker, self).__init__(*args, **kwargs)
        self.branch_commit_cache = {}

    def current_source_key(self, match):
        branch_key = (match.repo, match.branch)
        if branch_key in self.branch_commit_cache:
            key = self.branch_commit_cache[branch_key]
            match.branch_commit = key.commit
            return key

        commit = source.get_branch_commit(
            match.repo, match.branch)
        github_key = PyGithubKey(repo=match.repo, commit=commit)

        self.branch_commit_cache[(match.repo, match.branch)] = github_key
        match.branch_commit = commit[:8]

        return github_key

    def match_source_key(self, match):
        return PyGithubKey(repo=match.repo, commit=match.commit)

    def get_source(self, key, files):
        return source.download_files_from_repo(
            repo=key.repo, commit=key.commit, files=files)


class PyGithubMatch(PythonMatch):

    def __init__(self, filename, lineno, repo, path, element="",
                 branch="master"):
        try:
            self.repo, self.commit = repo.strip().split("@")
        except ValueError:
            raise NotMatching
        self.path = path.strip()
        self.element = element.strip()
        self.branch = branch

        # This may be filled manually later.
        self.branch_commit = None

        super(PyGithubMatch, self).__init__(filename, lineno)

    def __str__(self):
        return (
            "{match.repo}@{match.commit} vs {match.branch} branch"
            "{branch_commit} at {match.path}:{element} "
            "(from {match.filename}:{match.lineno})".format(
                match=self,
                branch_commit=" ({})".format(self.branch_commit)
                              if self.branch_commit else "",
                element=self.element or "whole module"))

    checker = PyGithubChecker
    match_type = "pygithub"
